#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include "../../util/cpp/graph/graph.h"

#define SPEED_OF_LIGHT     300000000 

FILE *fp;
graph_t G;
double X,Y,maxX,maxY,maxdist;

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s <graph-file> <tcl-file> <shift-x> <shift-y>\n", progname);
} // End usage()

void node_config(){

  Point p1,p2;
  double angle;
  FILE *fpr;
  if(!(fpr = fopen("relation.txt","w")))
    fprintf(stderr,"could not open file for writing \n");
  for (int i =0; i< G.numNodes();i++){
    int node_id = G.V[i].id;
    int no_edges = G.getDegree(node_id);
    G.getPoint(node_id,&p1);
    //  printf("node id %d degree %d x:%f y:%f \n",node_id,no_edges,p1.x,p1.y);
    fprintf(fp,"###### (%d) Node %d ###### \n",i,node_id);
    fprintf(fpr,"%d %d\n",i,node_id);
    fprintf(fp,"$ns_ node-config  -numif %d \n",no_edges);
    double x = (p1.x+X)*1000;
    double y = (p1.y+Y)*1000;
    if(x>maxX) maxX=x;
    if(y>maxY) maxY=y;
    if((x<0)) {fprintf(stderr,"X coordinates -ve, increase shift-x \n");exit(111);}
    if((y<0)) {fprintf(stderr,"Y coordinates -ve, increase shift-y \n");exit(111);}
    fprintf(fp,"set node_(%d) [create_node %f %f 0] \n",i,x,y);
    for(int j =0; j < no_edges;j++){
      int end_pt = G.V[i].nbrs[j].end_pt;
      G.getPoint(end_pt,&p2);
      double dist= p1.distance(p2)*1000;
      if (maxdist < dist) maxdist = dist;
      LineVec lv(p1,p2);
      angle = lv.angleWithXAxis();//in radians
      angle = angle * 180/M_PI; //in degree
      if(angle < 0) angle += 360;
      //printf("end pt %d x:%f y:%f angle %f;
      //",end_pt,p2.x,p2.y,angle);
      KeyValue_t kv(G.V[i].nbrs[j].s.cstr());
      const char *xstr = kv.get_value("txpower");
      float Pt = atof(xstr);
      KeyValue_t vv(G.V[i].nbrs[j].s.cstr());
      const char *ystr = vv.get_value("rcvth");
      double CST = atof(ystr);
      fprintf(fp,"### Interface %d (Node %d) ### \n",j,end_pt);
      fprintf(fp,"[$node_(%d) set netif_(%d)] set channel_number_ 1 \n",i,j);
      fprintf(fp,"[$node_(%d) set netif_(%d)] set Pt_ %f \n",i,j,Pt/1000);
      fprintf(fp,"[$node_(%d) set netif_(%d)] set CSThresh_ %e \n",i,j,CST/1000);
      //fprintf(fp,"[$node_(%d) set mac_(%d)] AirPropagationTime %0.6f \n",i,j,(ceil(dist*1000000/SPEED_OF_LIGHT)+2)/1000000);
      //fprintf(fp,"[$node_(%d) set mac_(%d)] AirPropagationConst 1 \n",i,j);
      fprintf(fp,"set a%d%d [new Antenna/DirAntenna] \n",i,j);
      fprintf(fp,"$a%d%d setType 1 \n",i,j);
      fprintf(fp,"$a%d%d setAngle %f \n",i,j,angle);
      fprintf(fp,"[$node_(%d) set netif_(%d)] dir-antenna $a%d%d \n",i,j,i,j);
    }//for j
    fprintf(fp,"\n");
  }//for i
}//node_config()

//returns the id of the node which is the landline
long landline(){
  for (int i =0; i< G.numNodes();i++){
    KeyValue_t kv(G.V[i].s.cstr());
    const char *xstr = kv.get_value("landline");
    if(xstr != NULL) return G.V[i].id;
  }
}//landline()

//The interface number between id1 and id2 is returned in ifa, 
//return -1 if id2 is not a neighbour of id1
int get_interface(long id1, long id2){
  int v_index = G.getNodeIndex(id1);
  for(int i=0; i< G.V[v_index].nbrs.size();i++){
    if(G.V[v_index].nbrs[i].end_pt == id2)
      return i;
  }
  return -1;
}//get_interface

void route_config(long lline) {
  vector<treeNode_t> tns;
  int ifa;
  G.shortest_path_tree(lline,tns);
                                                                                  
  fprintf(fp,"#addstaticroute <# hops> <next hop> <dest node> <ifa to use>#\n");
  
  fprintf(fp,"#### Routing-Uplink #### \n");
  for(int i=0; i<tns.size();i++){
    if(tns[i].id == lline) continue;//if landline, no up neighbours
    ifa = get_interface(tns[i].id,tns[i].parent);
    if(ifa==-1){
      fprintf(stderr,"the parent is not a neighbour!!!\n");
      exit(212);
    }
    fprintf(fp,"[$node_(%d) set ragent_] addstaticroute %d %d %d %d; ##%d:%d##\n",i,tns[i].depth,G.getNodeIndex(tns[i].parent),G.getNodeIndex(lline),ifa,tns[i].id,tns[i].parent);
  }// uplink
  
  fprintf(fp,"#### Routing-Downlink ####\n");
  for(int i=0; i<tns.size();i++){
    long parent = tns[i].id;
    int pid = G.getNodeIndex(parent);
    for(int j =0;j<tns[i].children.size();j++){
      long child = tns[i].children[j];
      int cid = G.getNodeIndex(child);
      int dist = tns[cid].depth-tns[pid].depth;
      if( dist == 1){ //immediate neighbours
	ifa = get_interface(parent,child);
	fprintf(fp,"[$node_(%d) set ragent_] addstaticroute %d %d %d %d; ##%d:%d##\n",pid,dist,cid,cid,ifa,parent,child);
	for(int k=0;k<tns[cid].children.size();k++){//grand children
	  long gchild = tns[cid].children[k];
	  int gid = G.getNodeIndex(gchild);
	  dist = tns[gid].depth-tns[pid].depth;
	  fprintf(fp,"[$node_(%d) set ragent_] addstaticroute %d %d %d %d; ##%d:%d##\n",pid,dist,cid,gid,ifa,parent,gchild);
	}
      }//if immediate neighbours
    }//for j
      
  }// for i downlink

}//route_config()


void arp_config(long lline) {
  vector<treeNode_t> tns;
  int ifa_child,ifa_parent,parent_ind,child_ind;
  G.shortest_path_tree(lline,tns);
                                                                                  
  fprintf(fp,"\n #add-arp-entry <node_addr> <mac_addr> #\n");
  
  for(int i=0; i<tns.size();i++){
    if(tns[i].id == lline) continue;//if landline, no up neighbours
    long child=tns[i].id;
    long parent = tns[i].parent; 
    child_ind = i;
    parent_ind = G.getNodeIndex(parent);
    ifa_child = get_interface(child,parent);
    if(ifa_child==-1){
      fprintf(stderr,"the parent is not a neighbour!!!\n");
      exit(212);
    }
    ifa_parent = get_interface(parent,child);
    if(ifa_parent==-1){
      fprintf(stderr,"the child is not a neighbour!!!\n");
      exit(212);
    }
    fprintf(fp,"set child_mac [[$node_(%d) set mac_(%d)] id];##%d:%d##\n",child_ind,ifa_child,child,parent);
    fprintf(fp,"set parent_mac [[$node_(%d) set mac_(%d)] id];##%d:%d##\n",parent_ind,ifa_parent,parent,child);
    fprintf(fp,"[$node_(%d) set ll_(%d)] add-arp-entry %d $parent_mac\n",child_ind,ifa_child,parent_ind);
    fprintf(fp,"[$node_(%d) set ll_(%d)] add-arp-entry %d $child_mac\n",parent_ind,ifa_parent,child_ind);
  }
  
}//arp_config()


int main(int argc, char *argv[]) {
  if(argc != 5) { usage(argv[0]); exit(1); }
  const char *gName = argv[1];
  const char *tName = argv[2];
  X = atof(argv[3]); 
  Y = atof(argv[4]);
  maxX=maxY=0;
  G.readGraph(gName);
  if(!(fp = fopen(tName,"w")))
    fprintf(stderr,"could not open file for writing \n");
  fprintf(fp,"set val(nn) %d; #number of mobile nodes\n",G.numNodes());
  fprintf(fp,"set val(lline) %d; #landline node\n",G.getNodeIndex(landline()));
  // Configure nodes in the network with interfaces & dir-antenna
  node_config();
  printf("landline is %d # nodes %d # edges %d maxX %f maxY %f maxprop %f\n",landline(),G.numNodes(),G.numEdges()/2,maxX,maxY,maxdist/SPEED_OF_LIGHT);
  route_config(landline());
  arp_config(landline());
  
}// End main()
