#include "wp2p-mac.h"
#include "wp2p-ll.h"


void WP2P_HandlerSend::handle(Event *e) {
  mac_->triggerSend();
} // End WP2P_HandlerSend::handle()

void WP2P_HandlerSend2Phy::handle(Event *e) {
  mac_->send2Phy((Packet*)e);
} // End WP2P_HandlerSend2Phy::handle()

void WP2P_HandlerEndRx::handle(Event *e){
  mac_->end_rx();
} // End WP2P_HandlerEndRx::handle()

void WP2P_HandlerEndTx::handle(Event *e){
  mac_->end_tx();
}//End WP2P_HandlerEndTx::handle()

void WP2P_HandlerTimeOut::handle(Event *e){
  mac_->timeout();
}//End WP2P_HandlerTimeOut::handle()

void WP2P_HandlerMinTimeInRx::handle(Event *e){
  mac_->done_min_time_in_rx();
}//End WP2P_HandlerMinTimeInRx::handle()

void WP2P_HandlerBumpTimeOut::handle(Event *e){
  mac_->bump_timeout();
}//End WP2P_HandlerBumpTimeOut::handle()


int hdr_mac2P::offset_;
static class Mac2PHeaderClass : public PacketHeaderClass {
public:
        Mac2PHeaderClass() : PacketHeaderClass("PacketHeader/Mac2P",
                                             sizeof(hdr_mac2P)) {
                bind_offset(&hdr_mac2P::offset_);
        }
} class_hdr_mac2P; 


static class WP2P_MACClass : public TclClass {
public:
  WP2P_MACClass() : TclClass("BiConnector/WP2P_MAC") {}
  TclObject* create(int, const char*const*) {
    return (new WP2P_MAC());
  }
} WP2P_class_MAC;

static int MacIndex = 0;

WP2P_MAC::WP2P_MAC() : BiConnector(), netif_(0), hSend(this), hSend2Phy(this), hEndRx(this), hEndTx(this), hTimeOut(this), hMinTimeInRx(this), hBumpTimeOut(this) {
  // Variable bindings
  index_ = MacIndex++;
  bind("send_n_", &send_n_);
  bind("recv_n_", &recv_n_);
  bind_bw("bandwidth_", &bandwidth_);
  bind_time("delay_", &delay_);

  // Variable initializations
  //  send_n_ = recv_n_ = 1;//CVR?? overwrite bindings?
  
  for(int i =0; i<MAX_IFA; i++){
    ifa_nbr_mac_[i] = NULL;
  }
  num_rx = 0;
  num_from_ll = 0;
  nbr_ifa = 0;

  RxTxState = IDLE; 
  SyncState = ASYNC;
  
  timer = -1;
  bump_timer = -1;
  
  timeout_count.setLimits(0, TIMEOUT_COUNT_MAX);
  async_count.setLimits(0, ASYNC_COUNT_MAX);
  bump_count.setLimits(0, BUMP_COUNT_MAX);
  rounds_since_prev_bump.setLimits(0, ROUNDS_BETN_BUMPS);
  try_bump = true;
  
} // End WP2P_MAC::WP2P_MAC()


int WP2P_MAC::command(int argc, const char*const* argv) {

  if(argc == 2) {
    if (strcmp(argv[1], "link-up") == 0) {
      link_up();
      return (TCL_OK);
    } else if (strcmp(argv[1], "link-down") ==0 ){
      link_down();
      return (TCL_OK);
    } else if(strcmp(argv[1], "id") == 0) {
      Tcl& tcl = Tcl::instance();
      tcl.resultf("%d", addr());
      return TCL_OK;
    } 
    
  }

  if (argc == 3) {
    if (strcmp(argv[1], "set-ifa-nbr") == 0) {
      ifa_nbr_mac_[nbr_ifa] = (WP2P_MAC*) TclObject::lookup(argv[2]);
      nbr_ifa++;
      return (TCL_OK);
    } else if (strcmp(argv[1], "set-ll") == 0) {
      ll_ = (WP2P_LL*) TclObject::lookup(argv[2]);
      return (TCL_OK);
    } else if (strcmp(argv[1],"set-delay")==0) {
      delay_ = atof(argv[2]);
      return (TCL_OK);
    } else if (strcmp(argv[1], "netif") == 0) {
      netif_ = (Phy*) TclObject::lookup(argv[2]);
      return TCL_OK;
    } else if (strcmp(argv[1], "log-target") == 0) {
      logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
      if(logtarget_ == 0)
	return TCL_ERROR;
      return TCL_OK;
    } 
    
  }
  return BiConnector::command(argc, argv);

} // End WP2P_MAC::command()

// pluck the precomputed tx time from the packet header
double WP2P_MAC::txtime(Packet *p) {
  struct hdr_cmn *ch = HDR_CMN(p);
  double t = ch->txtime();
  if (t < 0.0) {
    fprintf(stderr, "txtime is less than 0\n");
    exit(111);
  }
  return t;
}
 
//calculate tx time for packet of size "psz" bytes at rate "drt" bps
double WP2P_MAC::txtime(double psz, double drt) {
  double dsz = psz - PLCP_HDR_LEN;
  int plcp_hdr = PLCP_HDR_LEN << 3;
  int datalen = (int)dsz << 3;
  double t = (((double)plcp_hdr)/DSSS_PLCPDataRate) + (((double)datalen)/drt);
  return(t);
}


void WP2P_MAC::link_up() {
  fprintf(stderr, "%0.6f %d:%d LU MAC %c \n",
	  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
      
  //sanity check
  if((RxTxState != IDLE) || (SyncState != ASYNC)){
    fprintf(stderr,"Invalid state during link up of %d:%d MAC\n",ll_->my_node_id(),ll_->nbr_node_id());
    exit(149);
  }
  
  //initialization
  min_time_in_rx_done = true;
  timer = -1;
  bump_timer = -1;

  waiting_for_rx = false;
  timeout_count.set2lower();
  async_count.set2lower();
  bump_count.set2lower();
  rounds_since_prev_bump.set2lower();
  try_bump = true;
  
  /** Check the state of the other interfaces at the node. If some other
      interface is already up, copy its behaviour **/
  for(int i = 0; i < nbr_ifa; i++) {
    if(ifa_nbr_mac_[i]->RxTxState != IDLE){
      RxTxState = ifa_nbr_mac_[i]->RxTxState;
      return;
    }
  }

  /** If this is the first interface to be up, start tx on this
      interface **/
  update_tx_state();
  tx_on_link();
} // End WP2P_MAC::link_up()

void WP2P_MAC::link_down() {
  fprintf(stderr, "%0.6f %d:%d LD MAC %c \n",
	  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
  
  RxTxState = IDLE;
  SyncState = ASYNC;

} // End WP2P_MAC::link_down()



void WP2P_MAC::update_tx_state(){
  // Change to INSYNC, if in SYNCING
  // For each link, check timeout_count, async_count
  
    
  if(SyncState == SYNCING) {
    SyncState = INSYNC;
  }
  if(timeout_count.isUpper()) {
    timeout_count.set2lower();
    SyncState = ASYNC;
  }
  if(SyncState == ASYNC) {
    async_count.incr();
  } else {
    async_count.set2lower();
  }
  
  RxTxState = SYNTX;
  num_rx = 0;

}//End WP2P_MAC::update_tx_state() 


void WP2P_MAC::tx_on_all_links() {
  //update tx state of myself and my neighbours
  if(RxTxState!=IDLE) update_tx_state();
  for(int i = 0; i < nbr_ifa; i++) {
    if(ifa_nbr_mac_[i]->RxTxState != IDLE){
      ifa_nbr_mac_[i]->update_tx_state();
    }
  }
  //schedule tx on my link and neighbour interfaces
  if(RxTxState != IDLE) tx_on_link();
  for(int i = 0; i < nbr_ifa; i++) {
    if(ifa_nbr_mac_[i]->RxTxState != IDLE){
      ifa_nbr_mac_[i]->tx_on_link();
    }
  }
} // End WP2P_MAC::tx_on_all_links(

void WP2P_MAC::tx_on_link() {
  Scheduler &s = Scheduler::instance();
  s.schedule(&hSend, &hSendEvent,0);
} // WP2P_MAC::tx_on_link()

bool WP2P_MAC::ToBump() {
  // Determine if we need to bump, based on async_count  
  bool toBump = false;
  if(SyncState != ASYNC)
    return toBump;
  
  if(!try_bump) {
    // Need to increase rounds_since_prev_bump
    rounds_since_prev_bump.incr();
    if(rounds_since_prev_bump.isUpper()) {
      try_bump = true;
      rounds_since_prev_bump.set2lower();
      // can reset async_count also -- no need to bump now
      async_count.set2lower();
    }
  }
  
  if(try_bump){
    if(async_count.isUpper()) {
      async_count.set2lower();
      // Too many bumps?  Set link to IDLE state and retry syncing
      // after ROUNDS_BETN_BUMPS rounds
      bump_count.incr();
      if(bump_count.isUpper()) {
	bump_count.set2lower();
	try_bump = false;
	rounds_since_prev_bump.set2lower(); // start counting
      }
      if(try_bump) {
	toBump = true;
	fprintf(stderr, "%0.6f %d:%d AHB MAC %c\n",
		Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
      }
    }
  }
  return toBump;
}

void WP2P_MAC::checkBump(){
  bool bump = ToBump();
  for(int i = 0; i < nbr_ifa; i++) {
    if(ifa_nbr_mac_[i]->RxTxState != IDLE){
      if(ifa_nbr_mac_[i]->ToBump()==true)
	bump = true;
    }
  }
  
  if(bump==true){
    double dur = scheduleBumpTimer(-1);
    for(int i = 0; i < nbr_ifa; i++) {
      if(ifa_nbr_mac_[i]->RxTxState != IDLE)
	ifa_nbr_mac_[i]->scheduleBumpTimer(dur);
    }
  }
  else{
    tx_on_all_links();
  }
}

double WP2P_MAC::scheduleBumpTimer(double dur) {
  bump_timer = 1;
  if(dur == -1)
    dur = (double) rand()*BUMP_DUR/((double)RAND_MAX+1);
  fprintf(stderr,"bumping for %f \n",dur);
  Scheduler &s = Scheduler::instance();
  s.schedule(&hBumpTimeOut,&hBumpTimeOutEvent,dur);
  return dur;
}

void WP2P_MAC::cancelBumpTimer() {
  if(bump_timer != 1) {
    fprintf(stderr,"Error: bump timer is not active, exiting \n");
    exit(34);
  }
  bump_timer = -1;
  Scheduler &s = Scheduler::instance();
  s.cancel(&hBumpTimeOutEvent);
}
void WP2P_MAC::cancelMTRDTimer() {
  if(min_time_in_rx_done) {
    fprintf(stderr,"Error: MTRD timer is not active, exiting \n");
    exit(34);
  }
  min_time_in_rx_done = true;
  Scheduler &s = Scheduler::instance();
  s.cancel(&hMinTimeInRxEvent);
}

void WP2P_MAC::sendUp(Packet *p, Handler *h) {
  // We just received the 1st bit of a packet on the network
  // interface.
  
  struct hdr_cmn *hdr = HDR_CMN(p);
  hdr_mac2P *mh = HDR_MAC2P(p);
  u_int32_t dst = ETHER_ADDR(mh->dh_da);
  u_int32_t src = ETHER_ADDR(mh->dh_sa);

  if (RxTxState == IDLE){
    // The link can't see this pkt as its idle
    fprintf(stderr,"%0.6f %d:%d RxD MAC %c \n", 
	    Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
    Packet::free(p);
    return;
  }

  if(receiving == true){
    fprintf(stderr,"receiving again \n");
    if(dbg_recv == true){
      fprintf(stderr,"debug receiving again \n");
      int from_nbr = 0;
      for(int j = 0; j < nbr_ifa; j++) {
	if(src == (u_int32_t)ifa_nbr_mac_[j]->index_) {
	  from_nbr = 1; break;
	}
      }
      if (!from_nbr){
	fprintf(stderr,"%0.6f %d:%d Recv another pkt while already receiving one: Something is wrong\n",Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
	//exit(199);
      }
    }
    fprintf(stderr,"dropping pkt \n");
    fprintf(stderr,"%0.6f %d:%d RxD MAC %c \n",Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
    Packet::free(p);
    return;
  }
  
  receiving = true;
  dbg_recv = true;
  for(int j = 0; j < nbr_ifa; j++) {
    if(src == (u_int32_t)ifa_nbr_mac_[j]->index_) {
      dbg_recv = false; break;
    }
  }
  
  pktRx_ = p;
  fprintf(stderr, "%0.6f %d:%d RxS MAC %c\n",
	  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
  //If interface in TX mode, this packet won't be seen. So mark it as
  //errored, will be discarded later
  if(RxTxState == SYNTX ){
    //fprintf(stderr," Recv pkt when in TX mode \n");
    pktErr_ = 1;
  }
  
  if((RxTxState == SYNRX) && (dst == (u_int32_t)index_)) {
    
    if(bump_timer ==1){
      //Recv packet during bump_dur
      //fprintf(stderr,"%0.6f %d:%d recv packet during bumping \n",Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
      num_rx = 0;
      //cancel bump timers
      cancelBumpTimer();
      for(int i = 0; i < nbr_ifa; i++) {
	if(ifa_nbr_mac_[i]->RxTxState != IDLE)
	  ifa_nbr_mac_[i]->cancelBumpTimer();
      }
    }
  
    //CVR?? look into this iface thing
    /* BEGIN<Sabya,Ashwini> */
    hdr->iface()=addr();//...changed deliberately by us to revert bak to original
    /* END<Sabya,Ashwini> */
    
    /** If ASYNC, change to SYNCING -- will be changed to INSYNC on
	changing to SYNTX mode.  Need to check if no other link at node
	is also SYNCING. This is to ensure that only one interface is
	synching at any time **/
    bool to_sync = true;
    if(SyncState != ASYNC) { to_sync = false; }
    for(int j = 0; j < nbr_ifa; j++) {
      if(ifa_nbr_mac_[j]->SyncState == SYNCING) {
	to_sync = false; break;
      }
    }
    if(to_sync) {
      SyncState = SYNCING;
    }

    // Cancel timer
    if(timer != -1) {
      //fprintf(stderr, "l-d %0.6f %d:%d packet for me, cancel timer \n",
      //      Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
      Scheduler &s = Scheduler::instance();
      s.cancel(&hTimeOutEvent);
      timer = -1;
      waiting_for_rx = false;
      timeout_count.set2lower(); // reset counter
    }
  }//if SYNRX && dst
  //fprintf(stderr, "l-d %0.6f %d:%d schedule end of rx event\n",
  //  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
  Scheduler &s = Scheduler::instance();
  s.schedule(&hEndRx, (Event*)p, txtime(p));
}

void WP2P_MAC::end_rx() {
  hdr_mac2P *mh = HDR_MAC2P(pktRx_);
  u_int32_t dst = ETHER_ADDR(mh->dh_da);

  if (RxTxState == IDLE){
    // The link can't see this pkt as its idle
    fprintf(stderr,"%0.6f %d:%d RxD MAC %c \n", 
	    Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
    Packet::free(pktRx_);
    pktRx_=0;
    pktErr_=0;
    return;
  }
  receiving = false;
  dbg_recv = false;
  if(dst != (u_int32_t)index_) {
    fprintf(stderr, "%0.6f %d:%d RxE MAC\n", Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
    fprintf(stderr, "%0.6f %d:%d RxD MAC\n", Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
    Packet::free(pktRx_);
    pktRx_=0;
    pktErr_=0;
    return;
  }
  //If interface in TX mode, this packet won't be seen. So mark it as errored
  if(RxTxState == SYNTX){
    fprintf(stderr,"%0.6f %d:%d RxE MAC %c \n", Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
    fprintf(stderr, "%0.6f %d:%d RxD MAC\n", Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
    Packet::free(pktRx_);
    pktRx_=0;
    pktErr_=0;
    return;
  }
  
  //If packet was in error, it is dropped
  if(pktErr_==1){
    fprintf(stderr, "%0.6f %d:%d RxE MAC %c\n",
	    Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
    fprintf(stderr, "%0.6f %d:%d RxD MAC\n", Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
    Packet::free(pktRx_);
    pktRx_=0;
    pktErr_=0;
    return;
  }
  /** Simply send to LLC */
  fprintf(stderr, "%0.6f %d:%d RxE MAC %c\n",
	  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
  num_rx++;
  if(num_rx > recv_n_) {
    fprintf(stderr, "%0.6f %d:%d num_rx=%d, recv_n_=%d\n", Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(), num_rx, recv_n_); num_rx = recv_n_; 
  }
  //CVR?? h??
  //fprintf(stderr, "l-d %0.6f %d:%d send packet up to LL %d\n",
  //  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),num_rx);
  uptarget_->recv(pktRx_, (Handler*) 0);

  if(RxTxState == SYNRX) {
    // Check if links at node are done receiving
    // If so, start tx on all links
    if(all_done_receiving()) {
      checkBump();
    }
  } // End if(RxTxState == SYNRX)

} // End WP2P_MAC::end_rx()


bool WP2P_MAC::all_done_receiving() {
  //check my state
  if(SyncState != ASYNC){
    if(!min_time_in_rx_done) { return false;}
    if(receiving) {return false; }
    if(waiting_for_rx) {return false; }
    if(num_rx != recv_n_) {return false;}
  }
  // fprintf(stderr,"l-d %0.6f %d:%d I am done receiving \n",
  //  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());

  //check state of my neighbours
  for(int j = 0; j < nbr_ifa; j++) {
    if(ifa_nbr_mac_[j]->SyncState == ASYNC) { continue;}
    if(!ifa_nbr_mac_[j]->min_time_in_rx_done) {return false;}
    if(ifa_nbr_mac_[j]->receiving) { return false; }
    if(ifa_nbr_mac_[j]->waiting_for_rx) {  return false; }
    if(ifa_nbr_mac_[j]->num_rx != ifa_nbr_mac_[j]->recv_n_) {return false;}
  }
  //fprintf(stderr, "l-d %0.6f %d:%d All done receiving \n",
  // Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
  if(!min_time_in_rx_done){
    cancelMTRDTimer();
  }
  for(int j = 0; j < nbr_ifa; j++) {
    if(!ifa_nbr_mac_[j]->min_time_in_rx_done){
      ifa_nbr_mac_[j]->cancelMTRDTimer();
    }
  }
  if(bump_timer ==1){
    //cancel bump timers
    cancelBumpTimer();
    for(int i = 0; i < nbr_ifa; i++) {
      if(ifa_nbr_mac_[i]->RxTxState != IDLE)
	ifa_nbr_mac_[i]->cancelBumpTimer();
    }
  }
  return true;
} // End WP2P_MAC::all_done_receiving()


bool WP2P_MAC::end_tx() {
  // Change state to receive mode
  RxTxState = SYNRX;
  if(SyncState == ASYNC) {
    timer = -1; // no need to wait on this one
    waiting_for_rx = false;
  } else {
    // Need to wait to receive from INSYNC/SYNCING link
    waiting_for_rx = true;
    // Schedule timeout event in case the other side is down
    timer = 1; //timer is active
    Scheduler &s = Scheduler::instance();
    s.schedule(&hTimeOut,&hTimeOutEvent,TIMEOUT_VAL);
  }
  // Need to spend atleast PHASE_DUR+prop_delay+0.5*(R_MIN+R_MAX) in SYNRX
  //CVR? reason behind value?
  min_time_in_rx_done = false;
  fprintf(stderr, "%0.6f %d:%d TxE MAC %c\n", Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
  Scheduler &s = Scheduler::instance();
  s.schedule(&hMinTimeInRx,&hMinTimeInRxEvent,phase_dur+delay_+0.5*(R_MIN+R_MAX));
  return true;
} // End WP2P_MAC::end_tx()


bool WP2P_MAC::timeout() {
  timer = -1;
  waiting_for_rx = false; // don't wait anymore
  num_rx = recv_n_;
  timeout_count.incr();
  fprintf(stderr, "%0.6f %d:%d TO MAC %c\n",
	  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
  if(all_done_receiving()) {
    // all links are done receiving, start SYNTX on all links
    //fprintf(stderr, "l-d %0.6f %d:%d all done recv, tx on all links \n",
    //    Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
    checkBump();
  }
  return true;
} // End timeout()

void WP2P_MAC::bump_timeout() {
  //cancel bump timers
  fprintf(stderr, "%0.6f %d:%d TO MAC %c\n",
	  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
  cancelBumpTimer();
  for(int i = 0; i < nbr_ifa; i++) {
    if(ifa_nbr_mac_[i]->RxTxState != IDLE)
      ifa_nbr_mac_[i]->cancelBumpTimer();
  }
  if(!all_done_receiving()) {
    fprintf(stderr, "Error %0.6f %d:%d bump timeout but not all_done_recv \n",
	    Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
    exit(67);
  }
  tx_on_all_links();
} // End bump_timeout()


bool WP2P_MAC::done_min_time_in_rx() {
  fprintf(stderr, "%0.6f %d:%d MTRD MAC %c\n",
	  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
  min_time_in_rx_done = true;
  if(all_done_receiving()) {
    // all links are done receiving, start SYNTX on all links
    checkBump();
  }
  return true;
} // End WP2P_MAC::done_min_time_in_rx()

void WP2P_MAC::sendDown(Packet *p, Handler *h) {
  
  
  //  fprintf(stderr, "l-d %0.6f %d:%d recv packet %d from LL \n",
  //  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),num_from_ll);
  
  hdr_cmn* ch = HDR_CMN(p);
  fprintf(stderr,"MAC: recv pkt from higher layers of size %d \n",ch->size());  
  /* store data tx time */
  ch->txtime() = txtime(ch->size(), bandwidth_);
  //  fprintf(stderr," Pkt size %f txtime %f \n",(double)ch->size(),(double)ch->txtime());
	  
  // Just store the given packet in the array
  if(num_from_ll >= send_n_) {
    fprintf(stderr, "num_from_ll=%d, send_n_=%d\n", num_from_ll, send_n_); exit(8);
  }
  P[num_from_ll] = p;
  num_from_ll++;

  //fprintf(stderr, "mac: sendDown, num_from_ll=%d\n", num_from_ll);
  if(num_from_ll == send_n_){
    sendAll();
    num_from_ll = 0;
  }
} // End WP2P_MAC::sendDown()


void WP2P_MAC::triggerSend() {
  // In order to send, get packets from LLC
  //fprintf(stderr, "l-d %0.6f %d:%d ask LL for data \n",
  //    Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
  ll_->processSendData();
} // End WP2P_MAC::triggerSend()


void WP2P_MAC::sendAll() {
  
  //fprintf(stderr, "l-d %0.6f %d:%d recv req. pkts from LL \n",
  //  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id());
  double now = Scheduler::instance().clock();
  //fprintf(stderr, "%0.6f %d:%d MAC send-start\n", now, ll_->my_id(), ll_->nbr_id());
  // Just verify that all pkts have the same size

  for(int i = 0; i < num_from_ll; i++) {
    if(HDR_CMN(P[0])->size() != HDR_CMN(P[i])->size()) {
      fprintf(stderr, "size check failed\n"); exit(9);
    }
  }
  
  double txt = txtime(P[0]); // The transmission time of a single pkt
    // Schedule a send event with the downtarget_, for each of the
  // packet, spaced out within 'txt' of each other
  Scheduler& s = Scheduler::instance();
  fprintf(stderr, "%0.6f %d:%d TxS MAC %c\n",
	  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),(char)SyncState);
  for(int i = 0; i < num_from_ll; i++) {
    //fprintf(stderr, "l-d %0.6f %d:%d send event to downtarget %f \n",
    //    Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),i*txt);
    s.schedule(&hSend2Phy, (Event*)(P[i]), i*(txt+TX_WAIT_DUR));
  }
  //fprintf(stderr, "l-d %0.6f %d:%d finish tx event %f \n",
  //  Scheduler::instance().clock(), ll_->my_node_id(), ll_->nbr_node_id(),num_from_ll*txt);
  s.schedule(&hEndTx,&hEndTxEvent,(num_from_ll*txt+(num_from_ll-1)*TX_WAIT_DUR));
  phase_dur = num_from_ll*txt;
} // End WP2P_MAC::sendAll()


void WP2P_MAC::send2Phy(Packet *p) {
  downtarget_->recv(p, (Handler *)NULL);
} // End WP2P_MAC::send2Phy()
