/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* Ported from CMU/Monarch's code, nov'98 -Padma.*/

#include <math.h>
#include <stdlib.h>
#include <random.h>
#include <modulation.h>

/* BEGIN<Sabya,Ashwini> */

double get_interpolated_err(int x1,int x2,double y1,double y2,double x)
{
	return ((((x-x1)*(y2-y1))/(x2-x1))+ y1);
}	

/* ======================================================================
   Binary Phase Shift Keying
   ====================================================================== */

// Implementing the standard Bit Error Rate (BER) for DSSS PHY (2.4 GHz)
// 
// The function requires a table lookup (involves an integral that does not
// have a closed form)

BPSK::BPSK()
{
	Rs = 0;
	error_func_arr[0] = 0.0;
	error_func_arr[1] = 0.05637;
	error_func_arr[2] = 0.11246;
	error_func_arr[3] = 0.16800;
	error_func_arr[4] = 0.22270;
	error_func_arr[5] = 0.27633;
	error_func_arr[6] = 0.32863;
	error_func_arr[7] = 0.37938;
	error_func_arr[8] = 0.42839;
	error_func_arr[9] = 0.47548;
	error_func_arr[10] = 0.52050;
	error_func_arr[11] = 0.56332;	
	error_func_arr[12] = 0.60386;
	error_func_arr[13] = 0.64203;
	error_func_arr[14] = 0.67780;
	error_func_arr[15] = 0.71116;
	error_func_arr[16] = 0.74210;
	error_func_arr[17] = 0.77067;
	error_func_arr[18] = 0.79691;
	error_func_arr[19] = 0.82089;
	error_func_arr[20] = 0.84270;
	error_func_arr[21] = 0.86240;
	error_func_arr[22] = 0.88021;
	error_func_arr[23] = 0.89612;
	error_func_arr[24] = 0.91031;
	error_func_arr[25] = 0.92290;
	error_func_arr[26] = 0.93401;
	error_func_arr[27] = 0.94376;
	error_func_arr[28] = 0.95229;
	error_func_arr[29] = 0.95970;	
	error_func_arr[30] = 0.96611;
	error_func_arr[31] = 0.97162;
	error_func_arr[32] = 0.97635;
	error_func_arr[33] = 0.98038;
	error_func_arr[34] = 0.98379;
	error_func_arr[35] = 0.98667;
	error_func_arr[36] = 0.98909;
	error_func_arr[37] = 0.99111;
	error_func_arr[38] = 0.99279;
	error_func_arr[39] = 0.99418;
	error_func_arr[40] = 0.99532;
	error_func_arr[41] = 0.99959;
	error_func_arr[42] = 0.99998;
	error_func_arr[43] = 0.999998;
}

BPSK::BPSK(int S)
{
	Rs = S;
	error_func_arr[0] = 0.0;
	error_func_arr[1] = 0.05637;
	error_func_arr[2] = 0.11246;
	error_func_arr[3] = 0.16800;
	error_func_arr[4] = 0.22270;
	error_func_arr[5] = 0.27633;
	error_func_arr[6] = 0.32863;
	error_func_arr[7] = 0.37938;
	error_func_arr[8] = 0.42839;
	error_func_arr[9] = 0.47548;
	error_func_arr[10] = 0.52050;
	error_func_arr[11] = 0.56332;	
	error_func_arr[12] = 0.60386;
	error_func_arr[13] = 0.64203;
	error_func_arr[14] = 0.67780;
	error_func_arr[15] = 0.71116;
	error_func_arr[16] = 0.74210;
	error_func_arr[17] = 0.77067;
	error_func_arr[18] = 0.79691;
	error_func_arr[19] = 0.82089;
	error_func_arr[20] = 0.84270;
	error_func_arr[21] = 0.86240;
	error_func_arr[22] = 0.88021;
	error_func_arr[23] = 0.89612;
	error_func_arr[24] = 0.91031;
	error_func_arr[25] = 0.92290;
	error_func_arr[26] = 0.93401;
	error_func_arr[27] = 0.94376;
	error_func_arr[28] = 0.95229;
	error_func_arr[29] = 0.95970;	
	error_func_arr[30] = 0.96611;
	error_func_arr[31] = 0.97162;
	error_func_arr[32] = 0.97635;
	error_func_arr[33] = 0.98038;
	error_func_arr[34] = 0.98379;
	error_func_arr[35] = 0.98667;
	error_func_arr[36] = 0.98909;
	error_func_arr[37] = 0.99111;
	error_func_arr[38] = 0.99279;
	error_func_arr[39] = 0.99418;
	error_func_arr[40] = 0.99532;
	error_func_arr[41] = 0.99959;
	error_func_arr[42] = 0.99998;
	error_func_arr[43] = 0.999998;
}

int
BPSK::BitError(double Pr)
{
	double Pe;			// probability of error
	double x;
	int nbit = 0;			// number of bit errors tolerated

	if(nbit == 0) {
		Pe = ProbBitError(Pr);
	}
	else {
		Pe = ProbBitError(Pr, nbit);
	}

	// quick check
	if(Pe == 0.0)
	{
		return 0;		// no bit errors
	}	

	// scale the error probabilty
	Pe *= 1500;
	

	
	x = (double)((Random::uniform()) * 1000);

	if(x < Pe)
	{	
		return 1;		// bit error
	}	
	else
	{
		return 0;		// no bit errors
	}	
}


double BPSK::ProbBitError(double Pr)
{
	// assuming noise value = -90 dbm therefore noise == 10^-12 watts
	// also interpolation is required for values falling in intermediate range.
	
	double noise = 1e-10;
	double SNR = Pr/noise;
	double arg_err_func = sqrt(SNR);
	double magnified_value = 20*arg_err_func;
	int index = (int)magnified_value;
	 
	if(index < 40)
	{
		double pr = get_interpolated_err(index,index+1,error_func_arr[index],error_func_arr[index+1],magnified_value);
		double x = ((0.5)*(1-pr));
	
		return ((0.5)*(1-pr));
		
	}
	if(index >=40 && index < 50)
	{
		double pr = get_interpolated_err(40,50,error_func_arr[40],error_func_arr[41],magnified_value);
		double x = ((0.5)*(1-pr));
	
	
		return ((0.5)*(1-pr));
	}
	if(index >=50 && index < 60)
	{
		double pr = get_interpolated_err(50,60,error_func_arr[41],error_func_arr[42],magnified_value);
		double x = ((0.5)*(1-pr));
	
	
		return ((0.5)*(1-pr));
	}
	if(index >=60 && index <= 66)
	{
		double pr = get_interpolated_err(60,66,error_func_arr[42],error_func_arr[43],magnified_value);
			double x = ((0.5)*(1-pr));
	
	
		return ((0.5)*(1-pr));
	}
	if(index > 66)
	{
		double x = 0;
		
		
		return 0;	//check here whether this should be zero or 1
	}
	
}
double
BPSK::ProbBitError(double, int)
{
	double Pe = 0.0;

	return Pe;
}

/* END<Sabya,Ashwini> */

		
